# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#
# Autogenerated by Thrift Compiler (0.9.2)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,slots,dynamic
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
from ttypes import *
from thrift.Thrift import TProcessor
from thrift.protocol.TBase import TBase, TExceptionBase


class Iface(object):
  def testVoid(self):
    """
    Prints "testVoid()" and returns nothing.
    """
    pass

  def testString(self, thing):
    """
    Prints 'testString("%s")' with thing as '%s'
    @param string thing - the string to print
    @return string - returns the string 'thing'

    Parameters:
     - thing
    """
    pass

  def testByte(self, thing):
    """
    Prints 'testByte("%d")' with thing as '%d'
    @param byte thing - the byte to print
    @return byte - returns the byte 'thing'

    Parameters:
     - thing
    """
    pass

  def testI32(self, thing):
    """
    Prints 'testI32("%d")' with thing as '%d'
    @param i32 thing - the i32 to print
    @return i32 - returns the i32 'thing'

    Parameters:
     - thing
    """
    pass

  def testI64(self, thing):
    """
    Prints 'testI64("%d")' with thing as '%d'
    @param i64 thing - the i64 to print
    @return i64 - returns the i64 'thing'

    Parameters:
     - thing
    """
    pass

  def testDouble(self, thing):
    """
    Prints 'testDouble("%f")' with thing as '%f'
    @param double thing - the double to print
    @return double - returns the double 'thing'

    Parameters:
     - thing
    """
    pass

  def testBinary(self, thing):
    """
    Prints 'testBinary("%s")' where '%s' is a hex-formatted string of thing's data
    @param binary  thing - the binary data to print
    @return binary  - returns the binary 'thing'

    Parameters:
     - thing
    """
    pass

  def testStruct(self, thing):
    """
    Prints 'testStruct("{%s}")' where thing has been formatted into a string of comma separated values
    @param Xtruct thing - the Xtruct to print
    @return Xtruct - returns the Xtruct 'thing'

    Parameters:
     - thing
    """
    pass

  def testNest(self, thing):
    """
    Prints 'testNest("{%s}")' where thing has been formatted into a string of the nested struct
    @param Xtruct2 thing - the Xtruct2 to print
    @return Xtruct2 - returns the Xtruct2 'thing'

    Parameters:
     - thing
    """
    pass

  def testMap(self, thing):
    """
    Prints 'testMap("{%s")' where thing has been formatted into a string of  'key => value' pairs
     separated by commas and new lines
    @param map<i32,i32> thing - the map<i32,i32> to print
    @return map<i32,i32> - returns the map<i32,i32> 'thing'

    Parameters:
     - thing
    """
    pass

  def testStringMap(self, thing):
    """
    Prints 'testStringMap("{%s}")' where thing has been formatted into a string of  'key => value' pairs
     separated by commas and new lines
    @param map<string,string> thing - the map<string,string> to print
    @return map<string,string> - returns the map<string,string> 'thing'

    Parameters:
     - thing
    """
    pass

  def testSet(self, thing):
    """
    Prints 'testSet("{%s}")' where thing has been formatted into a string of  values
     separated by commas and new lines
    @param set<i32> thing - the set<i32> to print
    @return set<i32> - returns the set<i32> 'thing'

    Parameters:
     - thing
    """
    pass

  def testList(self, thing):
    """
    Prints 'testList("{%s}")' where thing has been formatted into a string of  values
     separated by commas and new lines
    @param list<i32> thing - the list<i32> to print
    @return list<i32> - returns the list<i32> 'thing'

    Parameters:
     - thing
    """
    pass

  def testEnum(self, thing):
    """
    Prints 'testEnum("%d")' where thing has been formatted into it's numeric value
    @param Numberz thing - the Numberz to print
    @return Numberz - returns the Numberz 'thing'

    Parameters:
     - thing
    """
    pass

  def testTypedef(self, thing):
    """
    Prints 'testTypedef("%d")' with thing as '%d'
    @param UserId thing - the UserId to print
    @return UserId - returns the UserId 'thing'

    Parameters:
     - thing
    """
    pass

  def testMapMap(self, hello):
    """
    Prints 'testMapMap("%d")' with hello as '%d'
    @param i32 hello - the i32 to print
    @return map<i32,map<i32,i32>> - returns a dictionary with these values:
      {-4 => {-4 => -4, -3 => -3, -2 => -2, -1 => -1, }, 4 => {1 => 1, 2 => 2, 3 => 3, 4 => 4, }, }

    Parameters:
     - hello
    """
    pass

  def testInsanity(self, argument):
    """
    So you think you've got this all worked, out eh?

    Creates a the returned map with these values and prints it out:
      { 1 => { 2 => argument,
               3 => argument,
             },
        2 => { 6 => <empty Insanity struct>, },
      }
    @return map<UserId, map<Numberz,Insanity>> - a map with the above values

    Parameters:
     - argument
    """
    pass

  def testMulti(self, arg0, arg1, arg2, arg3, arg4, arg5):
    """
    Prints 'testMulti()'
    @param byte arg0 -
    @param i32 arg1 -
    @param i64 arg2 -
    @param map<i16, string> arg3 -
    @param Numberz arg4 -
    @param UserId arg5 -
    @return Xtruct - returns an Xtruct with string_thing = "Hello2, byte_thing = arg0, i32_thing = arg1
       and i64_thing = arg2

    Parameters:
     - arg0
     - arg1
     - arg2
     - arg3
     - arg4
     - arg5
    """
    pass

  def testException(self, arg):
    """
    Print 'testException(%s)' with arg as '%s'
    @param string arg - a string indication what type of exception to throw
    if arg == "Xception" throw Xception with errorCode = 1001 and message = arg
    elsen if arg == "TException" throw TException
    else do not throw anything

    Parameters:
     - arg
    """
    pass

  def testMultiException(self, arg0, arg1):
    """
    Print 'testMultiException(%s, %s)' with arg0 as '%s' and arg1 as '%s'
    @param string arg - a string indication what type of exception to throw
    if arg0 == "Xception" throw Xception with errorCode = 1001 and message = "This is an Xception"
    elsen if arg0 == "Xception2" throw Xception2 with errorCode = 2002 and message = "This is an Xception2"
    else do not throw anything
    @return Xtruct - an Xtruct with string_thing = arg1

    Parameters:
     - arg0
     - arg1
    """
    pass

  def testOneway(self, secondsToSleep):
    """
    Print 'testOneway(%d): Sleeping...' with secondsToSleep as '%d'
    sleep 'secondsToSleep'
    Print 'testOneway(%d): done sleeping!' with secondsToSleep as '%d'
    @param i32 secondsToSleep - the number of seconds to sleep

    Parameters:
     - secondsToSleep
    """
    pass


class Client(Iface):
  def __init__(self, iprot, oprot=None):
    self._iprot = self._oprot = iprot
    if oprot is not None:
      self._oprot = oprot
    self._seqid = 0

  def testVoid(self):
    """
    Prints "testVoid()" and returns nothing.
    """
    self.send_testVoid()
    self.recv_testVoid()

  def send_testVoid(self):
    self._oprot.writeMessageBegin('testVoid', TMessageType.CALL, self._seqid)
    args = testVoid_args()
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testVoid(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testVoid_result()
    result.read(iprot)
    iprot.readMessageEnd()
    return

  def testString(self, thing):
    """
    Prints 'testString("%s")' with thing as '%s'
    @param string thing - the string to print
    @return string - returns the string 'thing'

    Parameters:
     - thing
    """
    self.send_testString(thing)
    return self.recv_testString()

  def send_testString(self, thing):
    self._oprot.writeMessageBegin('testString', TMessageType.CALL, self._seqid)
    args = testString_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testString(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testString_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testString failed: unknown result");

  def testByte(self, thing):
    """
    Prints 'testByte("%d")' with thing as '%d'
    @param byte thing - the byte to print
    @return byte - returns the byte 'thing'

    Parameters:
     - thing
    """
    self.send_testByte(thing)
    return self.recv_testByte()

  def send_testByte(self, thing):
    self._oprot.writeMessageBegin('testByte', TMessageType.CALL, self._seqid)
    args = testByte_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testByte(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testByte_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testByte failed: unknown result");

  def testI32(self, thing):
    """
    Prints 'testI32("%d")' with thing as '%d'
    @param i32 thing - the i32 to print
    @return i32 - returns the i32 'thing'

    Parameters:
     - thing
    """
    self.send_testI32(thing)
    return self.recv_testI32()

  def send_testI32(self, thing):
    self._oprot.writeMessageBegin('testI32', TMessageType.CALL, self._seqid)
    args = testI32_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testI32(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testI32_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testI32 failed: unknown result");

  def testI64(self, thing):
    """
    Prints 'testI64("%d")' with thing as '%d'
    @param i64 thing - the i64 to print
    @return i64 - returns the i64 'thing'

    Parameters:
     - thing
    """
    self.send_testI64(thing)
    return self.recv_testI64()

  def send_testI64(self, thing):
    self._oprot.writeMessageBegin('testI64', TMessageType.CALL, self._seqid)
    args = testI64_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testI64(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testI64_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testI64 failed: unknown result");

  def testDouble(self, thing):
    """
    Prints 'testDouble("%f")' with thing as '%f'
    @param double thing - the double to print
    @return double - returns the double 'thing'

    Parameters:
     - thing
    """
    self.send_testDouble(thing)
    return self.recv_testDouble()

  def send_testDouble(self, thing):
    self._oprot.writeMessageBegin('testDouble', TMessageType.CALL, self._seqid)
    args = testDouble_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testDouble(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testDouble_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testDouble failed: unknown result");

  def testBinary(self, thing):
    """
    Prints 'testBinary("%s")' where '%s' is a hex-formatted string of thing's data
    @param binary  thing - the binary data to print
    @return binary  - returns the binary 'thing'

    Parameters:
     - thing
    """
    self.send_testBinary(thing)
    return self.recv_testBinary()

  def send_testBinary(self, thing):
    self._oprot.writeMessageBegin('testBinary', TMessageType.CALL, self._seqid)
    args = testBinary_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testBinary(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testBinary_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testBinary failed: unknown result");

  def testStruct(self, thing):
    """
    Prints 'testStruct("{%s}")' where thing has been formatted into a string of comma separated values
    @param Xtruct thing - the Xtruct to print
    @return Xtruct - returns the Xtruct 'thing'

    Parameters:
     - thing
    """
    self.send_testStruct(thing)
    return self.recv_testStruct()

  def send_testStruct(self, thing):
    self._oprot.writeMessageBegin('testStruct', TMessageType.CALL, self._seqid)
    args = testStruct_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testStruct(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testStruct_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testStruct failed: unknown result");

  def testNest(self, thing):
    """
    Prints 'testNest("{%s}")' where thing has been formatted into a string of the nested struct
    @param Xtruct2 thing - the Xtruct2 to print
    @return Xtruct2 - returns the Xtruct2 'thing'

    Parameters:
     - thing
    """
    self.send_testNest(thing)
    return self.recv_testNest()

  def send_testNest(self, thing):
    self._oprot.writeMessageBegin('testNest', TMessageType.CALL, self._seqid)
    args = testNest_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testNest(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testNest_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testNest failed: unknown result");

  def testMap(self, thing):
    """
    Prints 'testMap("{%s")' where thing has been formatted into a string of  'key => value' pairs
     separated by commas and new lines
    @param map<i32,i32> thing - the map<i32,i32> to print
    @return map<i32,i32> - returns the map<i32,i32> 'thing'

    Parameters:
     - thing
    """
    self.send_testMap(thing)
    return self.recv_testMap()

  def send_testMap(self, thing):
    self._oprot.writeMessageBegin('testMap', TMessageType.CALL, self._seqid)
    args = testMap_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testMap(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testMap_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testMap failed: unknown result");

  def testStringMap(self, thing):
    """
    Prints 'testStringMap("{%s}")' where thing has been formatted into a string of  'key => value' pairs
     separated by commas and new lines
    @param map<string,string> thing - the map<string,string> to print
    @return map<string,string> - returns the map<string,string> 'thing'

    Parameters:
     - thing
    """
    self.send_testStringMap(thing)
    return self.recv_testStringMap()

  def send_testStringMap(self, thing):
    self._oprot.writeMessageBegin('testStringMap', TMessageType.CALL, self._seqid)
    args = testStringMap_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testStringMap(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testStringMap_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testStringMap failed: unknown result");

  def testSet(self, thing):
    """
    Prints 'testSet("{%s}")' where thing has been formatted into a string of  values
     separated by commas and new lines
    @param set<i32> thing - the set<i32> to print
    @return set<i32> - returns the set<i32> 'thing'

    Parameters:
     - thing
    """
    self.send_testSet(thing)
    return self.recv_testSet()

  def send_testSet(self, thing):
    self._oprot.writeMessageBegin('testSet', TMessageType.CALL, self._seqid)
    args = testSet_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testSet(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testSet_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testSet failed: unknown result");

  def testList(self, thing):
    """
    Prints 'testList("{%s}")' where thing has been formatted into a string of  values
     separated by commas and new lines
    @param list<i32> thing - the list<i32> to print
    @return list<i32> - returns the list<i32> 'thing'

    Parameters:
     - thing
    """
    self.send_testList(thing)
    return self.recv_testList()

  def send_testList(self, thing):
    self._oprot.writeMessageBegin('testList', TMessageType.CALL, self._seqid)
    args = testList_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testList(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testList_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testList failed: unknown result");

  def testEnum(self, thing):
    """
    Prints 'testEnum("%d")' where thing has been formatted into it's numeric value
    @param Numberz thing - the Numberz to print
    @return Numberz - returns the Numberz 'thing'

    Parameters:
     - thing
    """
    self.send_testEnum(thing)
    return self.recv_testEnum()

  def send_testEnum(self, thing):
    self._oprot.writeMessageBegin('testEnum', TMessageType.CALL, self._seqid)
    args = testEnum_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testEnum(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testEnum_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testEnum failed: unknown result");

  def testTypedef(self, thing):
    """
    Prints 'testTypedef("%d")' with thing as '%d'
    @param UserId thing - the UserId to print
    @return UserId - returns the UserId 'thing'

    Parameters:
     - thing
    """
    self.send_testTypedef(thing)
    return self.recv_testTypedef()

  def send_testTypedef(self, thing):
    self._oprot.writeMessageBegin('testTypedef', TMessageType.CALL, self._seqid)
    args = testTypedef_args()
    args.thing = thing
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testTypedef(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testTypedef_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testTypedef failed: unknown result");

  def testMapMap(self, hello):
    """
    Prints 'testMapMap("%d")' with hello as '%d'
    @param i32 hello - the i32 to print
    @return map<i32,map<i32,i32>> - returns a dictionary with these values:
      {-4 => {-4 => -4, -3 => -3, -2 => -2, -1 => -1, }, 4 => {1 => 1, 2 => 2, 3 => 3, 4 => 4, }, }

    Parameters:
     - hello
    """
    self.send_testMapMap(hello)
    return self.recv_testMapMap()

  def send_testMapMap(self, hello):
    self._oprot.writeMessageBegin('testMapMap', TMessageType.CALL, self._seqid)
    args = testMapMap_args()
    args.hello = hello
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testMapMap(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testMapMap_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testMapMap failed: unknown result");

  def testInsanity(self, argument):
    """
    So you think you've got this all worked, out eh?

    Creates a the returned map with these values and prints it out:
      { 1 => { 2 => argument,
               3 => argument,
             },
        2 => { 6 => <empty Insanity struct>, },
      }
    @return map<UserId, map<Numberz,Insanity>> - a map with the above values

    Parameters:
     - argument
    """
    self.send_testInsanity(argument)
    return self.recv_testInsanity()

  def send_testInsanity(self, argument):
    self._oprot.writeMessageBegin('testInsanity', TMessageType.CALL, self._seqid)
    args = testInsanity_args()
    args.argument = argument
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testInsanity(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testInsanity_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testInsanity failed: unknown result");

  def testMulti(self, arg0, arg1, arg2, arg3, arg4, arg5):
    """
    Prints 'testMulti()'
    @param byte arg0 -
    @param i32 arg1 -
    @param i64 arg2 -
    @param map<i16, string> arg3 -
    @param Numberz arg4 -
    @param UserId arg5 -
    @return Xtruct - returns an Xtruct with string_thing = "Hello2, byte_thing = arg0, i32_thing = arg1
       and i64_thing = arg2

    Parameters:
     - arg0
     - arg1
     - arg2
     - arg3
     - arg4
     - arg5
    """
    self.send_testMulti(arg0, arg1, arg2, arg3, arg4, arg5)
    return self.recv_testMulti()

  def send_testMulti(self, arg0, arg1, arg2, arg3, arg4, arg5):
    self._oprot.writeMessageBegin('testMulti', TMessageType.CALL, self._seqid)
    args = testMulti_args()
    args.arg0 = arg0
    args.arg1 = arg1
    args.arg2 = arg2
    args.arg3 = arg3
    args.arg4 = arg4
    args.arg5 = arg5
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testMulti(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testMulti_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testMulti failed: unknown result");

  def testException(self, arg):
    """
    Print 'testException(%s)' with arg as '%s'
    @param string arg - a string indication what type of exception to throw
    if arg == "Xception" throw Xception with errorCode = 1001 and message = arg
    elsen if arg == "TException" throw TException
    else do not throw anything

    Parameters:
     - arg
    """
    self.send_testException(arg)
    self.recv_testException()

  def send_testException(self, arg):
    self._oprot.writeMessageBegin('testException', TMessageType.CALL, self._seqid)
    args = testException_args()
    args.arg = arg
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testException(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testException_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.err1 is not None:
      raise result.err1
    return

  def testMultiException(self, arg0, arg1):
    """
    Print 'testMultiException(%s, %s)' with arg0 as '%s' and arg1 as '%s'
    @param string arg - a string indication what type of exception to throw
    if arg0 == "Xception" throw Xception with errorCode = 1001 and message = "This is an Xception"
    elsen if arg0 == "Xception2" throw Xception2 with errorCode = 2002 and message = "This is an Xception2"
    else do not throw anything
    @return Xtruct - an Xtruct with string_thing = arg1

    Parameters:
     - arg0
     - arg1
    """
    self.send_testMultiException(arg0, arg1)
    return self.recv_testMultiException()

  def send_testMultiException(self, arg0, arg1):
    self._oprot.writeMessageBegin('testMultiException', TMessageType.CALL, self._seqid)
    args = testMultiException_args()
    args.arg0 = arg0
    args.arg1 = arg1
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_testMultiException(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = testMultiException_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.err1 is not None:
      raise result.err1
    if result.err2 is not None:
      raise result.err2
    raise TApplicationException(TApplicationException.MISSING_RESULT, "testMultiException failed: unknown result");

  def testOneway(self, secondsToSleep):
    """
    Print 'testOneway(%d): Sleeping...' with secondsToSleep as '%d'
    sleep 'secondsToSleep'
    Print 'testOneway(%d): done sleeping!' with secondsToSleep as '%d'
    @param i32 secondsToSleep - the number of seconds to sleep

    Parameters:
     - secondsToSleep
    """
    self.send_testOneway(secondsToSleep)

  def send_testOneway(self, secondsToSleep):
    self._oprot.writeMessageBegin('testOneway', TMessageType.ONEWAY, self._seqid)
    args = testOneway_args()
    args.secondsToSleep = secondsToSleep
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

class Processor(Iface, TProcessor):
  def __init__(self, handler):
    self._handler = handler
    self._processMap = {}
    self._processMap["testVoid"] = Processor.process_testVoid
    self._processMap["testString"] = Processor.process_testString
    self._processMap["testByte"] = Processor.process_testByte
    self._processMap["testI32"] = Processor.process_testI32
    self._processMap["testI64"] = Processor.process_testI64
    self._processMap["testDouble"] = Processor.process_testDouble
    self._processMap["testBinary"] = Processor.process_testBinary
    self._processMap["testStruct"] = Processor.process_testStruct
    self._processMap["testNest"] = Processor.process_testNest
    self._processMap["testMap"] = Processor.process_testMap
    self._processMap["testStringMap"] = Processor.process_testStringMap
    self._processMap["testSet"] = Processor.process_testSet
    self._processMap["testList"] = Processor.process_testList
    self._processMap["testEnum"] = Processor.process_testEnum
    self._processMap["testTypedef"] = Processor.process_testTypedef
    self._processMap["testMapMap"] = Processor.process_testMapMap
    self._processMap["testInsanity"] = Processor.process_testInsanity
    self._processMap["testMulti"] = Processor.process_testMulti
    self._processMap["testException"] = Processor.process_testException
    self._processMap["testMultiException"] = Processor.process_testMultiException
    self._processMap["testOneway"] = Processor.process_testOneway

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      self._processMap[name](self, seqid, iprot, oprot)
    return True

  def process_testVoid(self, seqid, iprot, oprot):
    args = testVoid_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testVoid_result()
    self._handler.testVoid()
    oprot.writeMessageBegin("testVoid", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testString(self, seqid, iprot, oprot):
    args = testString_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testString_result()
    result.success = self._handler.testString(args.thing)
    oprot.writeMessageBegin("testString", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testByte(self, seqid, iprot, oprot):
    args = testByte_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testByte_result()
    result.success = self._handler.testByte(args.thing)
    oprot.writeMessageBegin("testByte", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testI32(self, seqid, iprot, oprot):
    args = testI32_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testI32_result()
    result.success = self._handler.testI32(args.thing)
    oprot.writeMessageBegin("testI32", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testI64(self, seqid, iprot, oprot):
    args = testI64_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testI64_result()
    result.success = self._handler.testI64(args.thing)
    oprot.writeMessageBegin("testI64", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testDouble(self, seqid, iprot, oprot):
    args = testDouble_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testDouble_result()
    result.success = self._handler.testDouble(args.thing)
    oprot.writeMessageBegin("testDouble", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testBinary(self, seqid, iprot, oprot):
    args = testBinary_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testBinary_result()
    result.success = self._handler.testBinary(args.thing)
    oprot.writeMessageBegin("testBinary", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testStruct(self, seqid, iprot, oprot):
    args = testStruct_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testStruct_result()
    result.success = self._handler.testStruct(args.thing)
    oprot.writeMessageBegin("testStruct", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testNest(self, seqid, iprot, oprot):
    args = testNest_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testNest_result()
    result.success = self._handler.testNest(args.thing)
    oprot.writeMessageBegin("testNest", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testMap(self, seqid, iprot, oprot):
    args = testMap_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testMap_result()
    result.success = self._handler.testMap(args.thing)
    oprot.writeMessageBegin("testMap", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testStringMap(self, seqid, iprot, oprot):
    args = testStringMap_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testStringMap_result()
    result.success = self._handler.testStringMap(args.thing)
    oprot.writeMessageBegin("testStringMap", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testSet(self, seqid, iprot, oprot):
    args = testSet_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testSet_result()
    result.success = self._handler.testSet(args.thing)
    oprot.writeMessageBegin("testSet", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testList(self, seqid, iprot, oprot):
    args = testList_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testList_result()
    result.success = self._handler.testList(args.thing)
    oprot.writeMessageBegin("testList", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testEnum(self, seqid, iprot, oprot):
    args = testEnum_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testEnum_result()
    result.success = self._handler.testEnum(args.thing)
    oprot.writeMessageBegin("testEnum", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testTypedef(self, seqid, iprot, oprot):
    args = testTypedef_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testTypedef_result()
    result.success = self._handler.testTypedef(args.thing)
    oprot.writeMessageBegin("testTypedef", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testMapMap(self, seqid, iprot, oprot):
    args = testMapMap_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testMapMap_result()
    result.success = self._handler.testMapMap(args.hello)
    oprot.writeMessageBegin("testMapMap", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testInsanity(self, seqid, iprot, oprot):
    args = testInsanity_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testInsanity_result()
    result.success = self._handler.testInsanity(args.argument)
    oprot.writeMessageBegin("testInsanity", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testMulti(self, seqid, iprot, oprot):
    args = testMulti_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testMulti_result()
    result.success = self._handler.testMulti(args.arg0, args.arg1, args.arg2, args.arg3, args.arg4, args.arg5)
    oprot.writeMessageBegin("testMulti", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testException(self, seqid, iprot, oprot):
    args = testException_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testException_result()
    try:
      self._handler.testException(args.arg)
    except Xception, err1:
      result.err1 = err1
    oprot.writeMessageBegin("testException", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testMultiException(self, seqid, iprot, oprot):
    args = testMultiException_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = testMultiException_result()
    try:
      result.success = self._handler.testMultiException(args.arg0, args.arg1)
    except Xception, err1:
      result.err1 = err1
    except Xception2, err2:
      result.err2 = err2
    oprot.writeMessageBegin("testMultiException", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_testOneway(self, seqid, iprot, oprot):
    args = testOneway_args()
    args.read(iprot)
    iprot.readMessageEnd()
    self._handler.testOneway(args.secondsToSleep)
    return


# HELPER FUNCTIONS AND STRUCTURES

class testVoid_args(TBase):

  __slots__ = [ 
   ]

  thrift_spec = (
  )

  def __hash__(self):
    value = 17
    return value


class testVoid_result(TBase):

  __slots__ = [ 
   ]

  thrift_spec = (
  )

  def __hash__(self):
    value = 17
    return value


class testString_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testString_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.STRING, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testByte_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.BYTE, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testByte_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.BYTE, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testI32_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testI32_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.I32, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testI64_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I64, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testI64_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.I64, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testDouble_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.DOUBLE, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testDouble_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.DOUBLE, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testBinary_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testBinary_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.STRING, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testStruct_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'thing', (Xtruct, Xtruct.thrift_spec), None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testStruct_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.STRUCT, 'success', (Xtruct, Xtruct.thrift_spec), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testNest_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'thing', (Xtruct2, Xtruct2.thrift_spec), None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testNest_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.STRUCT, 'success', (Xtruct2, Xtruct2.thrift_spec), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testMap_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.MAP, 'thing', (TType.I32,None,TType.I32,None), None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testMap_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.MAP, 'success', (TType.I32,None,TType.I32,None), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testStringMap_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.MAP, 'thing', (TType.STRING,None,TType.STRING,None), None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testStringMap_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.MAP, 'success', (TType.STRING,None,TType.STRING,None), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testSet_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.SET, 'thing', (TType.I32,None), None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testSet_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.SET, 'success', (TType.I32,None), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testList_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'thing', (TType.I32,None), None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testList_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.LIST, 'success', (TType.I32,None), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testEnum_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testEnum_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.I32, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testTypedef_args(TBase):
  """
  Attributes:
   - thing
  """

  __slots__ = [ 
    'thing',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I64, 'thing', None, None, ), # 1
  )

  def __init__(self, thing=None,):
    self.thing = thing

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.thing)
    return value


class testTypedef_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.I64, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testMapMap_args(TBase):
  """
  Attributes:
   - hello
  """

  __slots__ = [ 
    'hello',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'hello', None, None, ), # 1
  )

  def __init__(self, hello=None,):
    self.hello = hello

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.hello)
    return value


class testMapMap_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.MAP, 'success', (TType.I32,None,TType.MAP,(TType.I32,None,TType.I32,None)), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testInsanity_args(TBase):
  """
  Attributes:
   - argument
  """

  __slots__ = [ 
    'argument',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'argument', (Insanity, Insanity.thrift_spec), None, ), # 1
  )

  def __init__(self, argument=None,):
    self.argument = argument

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.argument)
    return value


class testInsanity_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.MAP, 'success', (TType.I64,None,TType.MAP,(TType.I32,None,TType.STRUCT,(Insanity, Insanity.thrift_spec))), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testMulti_args(TBase):
  """
  Attributes:
   - arg0
   - arg1
   - arg2
   - arg3
   - arg4
   - arg5
  """

  __slots__ = [ 
    'arg0',
    'arg1',
    'arg2',
    'arg3',
    'arg4',
    'arg5',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.BYTE, 'arg0', None, None, ), # 1
    (2, TType.I32, 'arg1', None, None, ), # 2
    (3, TType.I64, 'arg2', None, None, ), # 3
    (4, TType.MAP, 'arg3', (TType.I16,None,TType.STRING,None), None, ), # 4
    (5, TType.I32, 'arg4', None, None, ), # 5
    (6, TType.I64, 'arg5', None, None, ), # 6
  )

  def __init__(self, arg0=None, arg1=None, arg2=None, arg3=None, arg4=None, arg5=None,):
    self.arg0 = arg0
    self.arg1 = arg1
    self.arg2 = arg2
    self.arg3 = arg3
    self.arg4 = arg4
    self.arg5 = arg5

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.arg0)
    value = (value * 31) ^ hash(self.arg1)
    value = (value * 31) ^ hash(self.arg2)
    value = (value * 31) ^ hash(self.arg3)
    value = (value * 31) ^ hash(self.arg4)
    value = (value * 31) ^ hash(self.arg5)
    return value


class testMulti_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.STRUCT, 'success', (Xtruct, Xtruct.thrift_spec), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value


class testException_args(TBase):
  """
  Attributes:
   - arg
  """

  __slots__ = [ 
    'arg',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'arg', None, None, ), # 1
  )

  def __init__(self, arg=None,):
    self.arg = arg

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.arg)
    return value


class testException_result(TBase):
  """
  Attributes:
   - err1
  """

  __slots__ = [ 
    'err1',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'err1', (Xception, Xception.thrift_spec), None, ), # 1
  )

  def __init__(self, err1=None,):
    self.err1 = err1

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.err1)
    return value


class testMultiException_args(TBase):
  """
  Attributes:
   - arg0
   - arg1
  """

  __slots__ = [ 
    'arg0',
    'arg1',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'arg0', None, None, ), # 1
    (2, TType.STRING, 'arg1', None, None, ), # 2
  )

  def __init__(self, arg0=None, arg1=None,):
    self.arg0 = arg0
    self.arg1 = arg1

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.arg0)
    value = (value * 31) ^ hash(self.arg1)
    return value


class testMultiException_result(TBase):
  """
  Attributes:
   - success
   - err1
   - err2
  """

  __slots__ = [ 
    'success',
    'err1',
    'err2',
   ]

  thrift_spec = (
    (0, TType.STRUCT, 'success', (Xtruct, Xtruct.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'err1', (Xception, Xception.thrift_spec), None, ), # 1
    (2, TType.STRUCT, 'err2', (Xception2, Xception2.thrift_spec), None, ), # 2
  )

  def __init__(self, success=None, err1=None, err2=None,):
    self.success = success
    self.err1 = err1
    self.err2 = err2

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.err1)
    value = (value * 31) ^ hash(self.err2)
    return value


class testOneway_args(TBase):
  """
  Attributes:
   - secondsToSleep
  """

  __slots__ = [ 
    'secondsToSleep',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'secondsToSleep', None, None, ), # 1
  )

  def __init__(self, secondsToSleep=None,):
    self.secondsToSleep = secondsToSleep

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.secondsToSleep)
    return value




from setuptools import find_packages, setup


setup(
    name='tchannel',
    version='0.17.3',
    author='Abhinav Gupta, Aiden Scandella, Bryce Lampe, Grayson Koonce, Junchao Wu',
    author_email='abg@uber.com',
    description='Network multiplexing and framing protocol for RPC',
    license='MIT',
    url='https://github.com/uber/tchannel-python',
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={
        '': ['*.thrift'],
    },
    install_requires=[
        'contextlib2',
        'crcmod',
        'tornado>=4.0,<5.0',
        'toro>=0.8,<0.9',
        'thriftrw>=0.3,<0.5',
        'threadloop>=0.5,<0.6',
        'futures',
    ],
    extras_require={
        'vcr': ['PyYAML', 'mock', 'wrapt'],
    },
    entry_points={
        'console_scripts': [
            'tcurl.py = tchannel.tcurl:start_ioloop'
        ]
    },
)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import sys
from collections import namedtuple

from tornado import gen

from tchannel.response import Response, response_from_mixed
from tchannel.tornado.request import TransportMetadata
from tchannel.tornado.response import StatusCode

from ..serializer.thrift import ThriftSerializer


def register(dispatcher, service_module, handler, method=None, service=None):
    """Registers a Thrift service method with the given RequestDispatcher.

    .. code-block:: python

        # For,
        #
        #   service HelloWorld { string hello(1: string name); }

        import tchannel.thrift
        import HelloWorld

        def hello(request, response):
            name = request.args.name
            response.write_result("Hello, %s" % name)

        dispatcher = RequestDispatcher()
        tchannel.thrift.register(dispatcher, HelloWorld, hello)

    :param dispatcher:
        TChannel dispatcher with which the Thrift service will be registered.
    :param service_module:
        The service module generated by Thrift. This module contains the
        service ``Iface``, ``Client``, ``Processor``, etc. classes.
    :param handler:
        A function implementing the request handler. The function must accept
        a ``request``, a ``response``, and a ``tchannel``.
    :param service:
        Thrift service name. This is the `service` name specified in the
        Thrift IDL. If omitted, it is automatically determined based on the
        name of ``service_module``.
    :param method:
        Name of the method. Defaults to the name of the ``handler`` function.
    """
    if not service:
        service = service_module.__name__.rsplit('.', 1)[-1]
    if not method:
        method = handler.__name__
    assert service, 'A service name could not be determined'
    assert method, 'A method name could not be determined'

    assert hasattr(service_module.Iface, method), (
        "Service %s doesn't define method %s" % (service, method)
    )
    assert hasattr(service_module, method + '_result'), (
        "oneway methods are not yet supported"
    )

    endpoint = '%s::%s' % (service, method)
    args_type = getattr(service_module, method + '_args')
    result_type = getattr(service_module, method + '_result')

    # if the dispatcher is set to deal with handlers that
    # return responses, then use new api, else use deprecated
    if dispatcher._handler_returns_response:
        handler = build_handler(result_type, handler)
    else:
        handler = deprecated_build_handler(result_type, handler)

    dispatcher.register(
        endpoint,
        handler,
        ThriftSerializer(args_type),
        ThriftSerializer(result_type)
    )
    return handler


def build_handler(result_type, f):
    @gen.coroutine
    def handler(request):

        result = ThriftResponse(result_type())
        response = Response()

        try:
            response = yield gen.maybe_future(f(request))
        except Exception:
            result.write_exc_info(sys.exc_info())
        else:
            response = response_from_mixed(response)
            result.write_result(response.body)

        response.status = result.code
        response.body = result.result

        raise gen.Return(response)
    return handler


def deprecated_build_handler(result_type, f):
    @gen.coroutine
    def handler(request, response):
        req = yield ThriftRequest._from_raw_request(request)
        res = ThriftResponse(result_type())
        try:
            # TODO: It would be nice if we could wait until write_result was
            # called or an exception was thrown instead of waiting for the
            # function to return. This would allow for use cases where the
            # implementation returns the result early but still does some work
            # after that.
            result = yield gen.maybe_future(f(req, res))
        except Exception:
            res.write_exc_info(sys.exc_info())
        else:
            if not res.finished and result is not None:
                # The user never called write_result or threw an
                # exception. The result was most likely returned by the
                # function.
                res.write_result(result)
        response.code = res.code
        response.write_header(res.headers)
        response.write_body(res.result)
    return handler


class ThriftRequest(namedtuple('_Request', 'headers args transport')):
    """Represents a Thrift call request.

    Makes the following attributes available:

    ``headers``
        The application-level headers passed in for this request.

    ``args``
        An object containing the parameters for the Thrift call. The object's
        attributes will have the same name as the parameters defined in the
        Thrift IDL.

    ``transport``
        Provides access to the transport metadata. Among other things, this
        includes,

        ``headers``
            The transport-level headers
        ``ttl``
            TTL for this request in milliseconds. The caller expects a
            response within this period. If this time is exceeded, the timeout
            may kick in at the call site or in the forwarding layer.
    """

    @classmethod
    @gen.coroutine
    def _from_raw_request(cls, request):
        call_headers = yield request.get_header()
        call_args = yield request.get_body()
        transport_metadata = TransportMetadata.from_request(request)
        raise gen.Return(
            cls(
                headers=call_headers,
                args=call_args,
                transport=transport_metadata,
            )
        )


class ThriftResponse(object):
    """Represents a response to a Thrift call."""

    __slots__ = ('headers', 'result', 'finished', 'code')

    def __init__(self, result=None):
        self.headers = {}
        self.result = result
        self.finished = False
        self.code = StatusCode.ok

    def write_header(self, name, value):
        """Add a header to be written in the response."""
        self.headers[name] = value

    def write_result(self, result):
        """Send back the result of this call.

        Only one of this and `write_exc_info` may be called.

        :param result:
            Return value of the call
        """
        assert not self.finished, "Already sent a response"

        if not self.result.thrift_spec:
            self.finished = True
            return

        spec = self.result.thrift_spec[0]
        if result is not None:
            assert spec, "Tried to return a result for a void method."
            setattr(self.result, spec[2], result)

        self.finished = True

    def write_exc_info(self, exc_info=None):
        """Write exception information to the response.

        Only one of this and ``write_result`` may be called.

        :param exc_info:
            3-tuple of exception information. If omitted, the last exception
            will be retrieved using ``sys.exc_info()``.
        """
        exc_info = exc_info or sys.exc_info()
        exc = exc_info[1]
        self.code = StatusCode.error
        for spec in self.result.thrift_spec[1:]:
            if spec and isinstance(exc, spec[3][0]):
                assert not self.finished, "Already sent a response"

                setattr(self.result, spec[2], exc)
                self.finished = True
                return

        # Re-raise the exception (with the same traceback) if it didn't match.
        raise exc_info[0], exc_info[1], exc_info[2]



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#
# Autogenerated by Thrift Compiler (0.9.2)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:tornado,dynamic,slots,dynexc=VCRThriftError,dynbase=VCRThriftBase,dynimport=from tchannel.testing.vcr.thrift import *
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
from ttypes import *
from thrift.Thrift import TProcessor
from tchannel.testing.vcr.thrift import *
from tornado import gen
from tornado import concurrent
from thrift.transport import TTransport

class Iface(object):
  """
  The VCRProxy service is responsible for forwarding requests to the remote
  server, recording the interactions, and replaying responses for known
  requests.
  """
  def send(self, request):
    """
    Send the given request through the system.

    If the request is known, replay its response. Otherwise, forward it to
    the remote server and return the remote server's response.

    Parameters:
     - request
    """
    pass


class Client(Iface):
  """
  The VCRProxy service is responsible for forwarding requests to the remote
  server, recording the interactions, and replaying responses for known
  requests.
  """
  def __init__(self, transport, iprot_factory, oprot_factory=None):
    self._transport = transport
    self._iprot_factory = iprot_factory
    self._oprot_factory = (oprot_factory if oprot_factory is not None
                           else iprot_factory)
    self._seqid = 0
    self._reqs = {}
    self._transport.io_loop.spawn_callback(self._start_receiving)

  @gen.engine
  def _start_receiving(self):
    while True:
      try:
        frame = yield self._transport.readFrame()
      except TTransport.TTransportException as e:
        for future in self._reqs.itervalues():
          future.set_exception(e)
        self._reqs = {}
        return
      tr = TTransport.TMemoryBuffer(frame)
      iprot = self._iprot_factory.getProtocol(tr)
      (fname, mtype, rseqid) = iprot.readMessageBegin()
      future = self._reqs.pop(rseqid, None)
      if not future:
        # future has already been discarded
        continue
      method = getattr(self, 'recv_' + fname)
      try:
        result = method(iprot, mtype, rseqid)
      except Exception as e:
        future.set_exception(e)
      else:
        future.set_result(result)

  def send(self, request):
    """
    Send the given request through the system.

    If the request is known, replay its response. Otherwise, forward it to
    the remote server and return the remote server's response.

    Parameters:
     - request
    """
    self._seqid += 1
    future = self._reqs[self._seqid] = concurrent.Future()
    self.send_send(request)
    return future

  def send_send(self, request):
    oprot = self._oprot_factory.getProtocol(self._transport)
    oprot.writeMessageBegin('send', TMessageType.CALL, self._seqid)
    args = send_args()
    args.request = request
    args.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def recv_send(self, iprot, mtype, rseqid):
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = send_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.cannotRecord is not None:
      raise result.cannotRecord
    if result.remoteServiceError is not None:
      raise result.remoteServiceError
    if result.serviceError is not None:
      raise result.serviceError
    if result.noPeersAvailable is not None:
      raise result.noPeersAvailable
    raise TApplicationException(TApplicationException.MISSING_RESULT, "send failed: unknown result");


class Processor(Iface, TProcessor):
  def __init__(self, handler):
    self._handler = handler
    self._processMap = {}
    self._processMap["send"] = Processor.process_send

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      return self._processMap[name](self, seqid, iprot, oprot)

  @gen.coroutine
  def process_send(self, seqid, iprot, oprot):
    args = send_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = send_result()
    try:
      result.success = yield gen.maybe_future(self._handler.send(args.request))
    except CannotRecordInteractionsError, cannotRecord:
      result.cannotRecord = cannotRecord
    except RemoteServiceError, remoteServiceError:
      result.remoteServiceError = remoteServiceError
    except VCRServiceError, serviceError:
      result.serviceError = serviceError
    except NoPeersAvailableError, noPeersAvailable:
      result.noPeersAvailable = noPeersAvailable
    oprot.writeMessageBegin("send", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES

class send_args(VCRThriftBase):
  """
  Attributes:
   - request
  """

  __slots__ = [ 
    'request',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'request', (Request, Request.thrift_spec), None, ), # 1
  )

  def __init__(self, request=None,):
    self.request = request

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.request)
    return value


class send_result(VCRThriftBase):
  """
  Attributes:
   - success
   - cannotRecord: Thrown if the request was unrecognized and the record mode for the
  current cassette disallows recording new interactions.
   - remoteServiceError
   - serviceError
   - noPeersAvailable
  """

  __slots__ = [ 
    'success',
    'cannotRecord',
    'remoteServiceError',
    'serviceError',
    'noPeersAvailable',
   ]

  thrift_spec = (
    (0, TType.STRUCT, 'success', (Response, Response.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'cannotRecord', (CannotRecordInteractionsError, CannotRecordInteractionsError.thrift_spec), None, ), # 1
    (2, TType.STRUCT, 'remoteServiceError', (RemoteServiceError, RemoteServiceError.thrift_spec), None, ), # 2
    (3, TType.STRUCT, 'serviceError', (VCRServiceError, VCRServiceError.thrift_spec), None, ), # 3
    (4, TType.STRUCT, 'noPeersAvailable', (NoPeersAvailableError, NoPeersAvailableError.thrift_spec), None, ), # 4
  )

  def __init__(self, success=None, cannotRecord=None, remoteServiceError=None, serviceError=None, noPeersAvailable=None,):
    self.success = success
    self.cannotRecord = cannotRecord
    self.remoteServiceError = remoteServiceError
    self.serviceError = serviceError
    self.noPeersAvailable = noPeersAvailable

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.cannotRecord)
    value = (value * 31) ^ hash(self.remoteServiceError)
    value = (value * 31) ^ hash(self.serviceError)
    value = (value * 31) ^ hash(self.noPeersAvailable)
    return value




# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import logging
from collections import namedtuple

import tornado
import tornado.gen
from tornado import gen

from tchannel import transport
from tchannel.request import Request
from tchannel.request import TransportHeaders
from tchannel.response import response_from_mixed
from ..context import request_context
from ..errors import BadRequestError
from ..errors import TChannelError
from ..event import EventType
from ..messages import Types
from ..messages.error import ErrorCode
from ..serializer.raw import RawSerializer
from .response import Response as DeprecatedResponse

log = logging.getLogger('tchannel')


Handler = namedtuple('Handler', 'endpoint req_serializer resp_serializer')


class RequestDispatcher(object):
    """A synchronous RequestHandler that dispatches calls to different
    endpoints based on ``arg1``.

    Endpoints are registered using ``register`` or the ``route``
    decorator.

    .. code-block:: python

        handler = # ...

        @handler.route('my_method')
        def my_method(request, response):
            response.write('hello world')
    """

    FALLBACK = object()

    def __init__(self, _handler_returns_response=False):
        self.handlers = {}
        self.register(self.FALLBACK, self.not_found)
        self._handler_returns_response = _handler_returns_response

    _HANDLER_NAMES = {
        Types.CALL_REQ: 'pre_call',
        Types.CALL_REQ_CONTINUE: 'pre_call'
    }

    def handle(self, message, connection):
        # TODO assert that the handshake was already completed
        assert message, "message must not be None"

        if message.message_type not in self._HANDLER_NAMES:
            # TODO handle this more gracefully
            raise NotImplementedError("Unexpected message: %s" % str(message))

        handler_name = "handle_" + self._HANDLER_NAMES[message.message_type]
        return getattr(self, handler_name)(message, connection)

    def handle_pre_call(self, message, connection):
        """Handle incoming request message including CallRequestMessage and
        CallRequestContinueMessage

        This method will build the User friendly request object based on the
        incoming messages.

        It passes all the messages into the message_factory to build the init
        request object. Only when it get a CallRequestMessage and a completed
        arg_1=argstream[0], the message_factory will return a request object.
        Then it will trigger the async call_handle call.

        :param message: CallRequestMessage or CallRequestContinueMessage
        :param connection: tornado connection
        """
        try:
            req = connection.request_message_factory.build(message)
            # message_factory will create Request only when it receives
            # CallRequestMessage. It will return None, if it receives
            # CallRequestContinueMessage.
            if req:
                self.handle_call(req, connection)

        except TChannelError as e:
            log.warn('Received a bad request.', exc_info=True)

            connection.send_error(
                e.code,
                e.message,
                message.id,
            )

    @tornado.gen.coroutine
    def handle_call(self, request, connection):
        # read arg_1 so that handle_call is able to get the endpoint
        # name and find the endpoint handler.
        # the arg_1 value will be store in the request.endpoint field.

        # NOTE: after here, the correct way to access value of arg_1 is through
        # request.endpoint. The original argstream[0] is no longer valid. If
        # user still tries read from it, it will return empty.
        chunk = yield request.argstreams[0].read()
        response = None
        while chunk:
            request.endpoint += chunk
            chunk = yield request.argstreams[0].read()

        log.debug('Received a call to %s.', request.endpoint)

        tchannel = connection.tchannel

        # event: receive_request
        request.tracing.name = request.endpoint
        tchannel.event_emitter.fire(EventType.before_receive_request, request)

        handler = self.handlers.get(request.endpoint)

        if handler is None:
            handler = self.handlers[self.FALLBACK]

        requested_as = request.headers.get('as', None)
        expected_as = handler.req_serializer.name

        if request.endpoint in self.handlers and requested_as != expected_as:
            connection.send_error(
                ErrorCode.bad_request,
                "Your serialization was '%s' but the server expected '%s'" % (
                    requested_as,
                    expected_as,
                ),
                request.id,
            )
            raise gen.Return(None)

        request.serializer = handler.req_serializer
        response = DeprecatedResponse(
            id=request.id,
            checksum=request.checksum,
            tracing=request.tracing,
            connection=connection,
            headers={'as': request.headers.get('as', 'raw')},
            serializer=handler.resp_serializer,
        )

        connection.post_response(response)

        try:
            # New impl - the handler takes a request and returns a response
            if self._handler_returns_response:

                # convert deprecated req to new top-level req
                b = yield request.get_body()
                he = yield request.get_header()
                t = request.headers
                t = transport.to_kwargs(t)
                t = TransportHeaders(**t)
                new_req = Request(
                    body=b,
                    headers=he,
                    transport=t,
                    endpoint=request.endpoint,
                )

                # Not safe to have coroutine yields statement within
                # stack context.
                # The right way to do it is:
                # with request_context(..):
                #    future = f()
                # yield future

                with request_context(request.tracing):
                    f = handler.endpoint(new_req)

                new_resp = yield gen.maybe_future(f)

                # instantiate a tchannel.Response
                new_resp = response_from_mixed(new_resp)

                response.code = new_resp.status

                # assign resp values to dep response
                response.write_header(new_resp.headers)

                if new_resp.body is not None:
                    response.write_body(new_resp.body)

            # Dep impl - the handler is provided with a req & resp writer
            else:
                with request_context(request.tracing):
                    f = handler.endpoint(request, response)

                yield gen.maybe_future(f)

            response.flush()
        except TChannelError as e:
            connection.send_error(
                e.code,
                e.message,
                request.id,
            )
        except Exception as e:
            msg = "An unexpected error has occurred from the handler"
            log.exception(msg)

            response.set_exception(TChannelError(e.message))
            connection.request_message_factory.remove_buffer(response.id)

            connection.send_error(ErrorCode.unexpected, msg, response.id)
            tchannel.event_emitter.fire(EventType.on_exception, request, e)

        raise gen.Return(response)

    def register(
            self,
            rule,
            handler,
            req_serializer=None,
            resp_serializer=None
    ):
        """Register a new endpoint with the given name.

        .. code-block:: python

            @dispatcher.register('is_healthy')
            def check_health(request, response):
                # ...

        :param rule:
            Name of the endpoint. Incoming Call Requests must have this as
            ``arg1`` to dispatch to this handler.

            If ``RequestHandler.FALLBACK`` is specified as a rule, the given
            handler will be used as the 'fallback' handler when requests don't
            match any registered rules.

        :param handler:
            A function that gets called with ``Request`` and ``Response``.

        :param req_serializer:
            Arg scheme serializer of this endpoint. It should be
            ``RawSerializer``, ``JsonSerializer``, and ``ThriftSerializer``.

        :param resp_serializer:
            Arg scheme serializer of this endpoint. It should be
            ``RawSerializer``, ``JsonSerializer``, and ``ThriftSerializer``.
        """

        assert handler, "handler must not be None"
        req_serializer = req_serializer or RawSerializer()
        resp_serializer = resp_serializer or RawSerializer()
        self.handlers[rule] = Handler(handler, req_serializer, resp_serializer)

    @staticmethod
    def not_found(request, response=None):
        """Default behavior for requests to unrecognized endpoints."""
        raise BadRequestError(
            description="Endpoint '%s' is not defined" % (
                request.endpoint,
            ),
        )
